from builtins import zip
__docformat__ = "restructuredtext en"

import mdp
from mdp import numx

from .svm_classifiers import _SVMClassifier, _LabelNormalizer

import svmutil as libsvmutil

class LibSVMClassifier(_SVMClassifier):
    """
    The ``LibSVMClassifier`` class acts as a wrapper around the LibSVM library
    for support vector machines.

    Information to the parameters can be found on
    http://www.csie.ntu.edu.tw/~cjlin/libsvm/
    
    The class provides access to change kernel and svm type with a text string.
    
    Additionally ``self.parameter`` is exposed which allows to change all other
    svm parameters directly.

    This node depends on ``libsvm``.
    """
    # The kernels and classifiers which LibSVM allows.
    kernels = ["RBF", "LINEAR", "POLY", "SIGMOID"]
    classifiers = ["C_SVC", "NU_SVC", "ONE_CLASS", "EPSILON_SVR", "NU_SVR"]

    def __init__(self, kernel=None, classifier=None, probability=True, params=None,
                 input_dim=None, output_dim=None, dtype=None):
        """
        kernel -- The kernel to use
        classifier -- The type of the SVM
        params -- a dict of parameters to be passed to the svm_parameter
        probability -- Must be set to True, if algorithms based on probability
                       shall be used.
        """
        if not params:
            params = {}
        
        # initialise the parameter and be quiet
        self.parameter = libsvmutil.svm_parameter("-q")
        if probability:
            # allow for probability estimates
            self.parameter.probability = 1
        
        super(LibSVMClassifier, self).__init__(input_dim=input_dim,
                                               output_dim=output_dim,
                                               dtype=dtype)
        if kernel:
            self.set_kernel(kernel)
        if classifier:
            self.set_classifier(classifier)
        # set all other parameters
        for k, v in params.items():
            if not k in self.parameter._names:
                # check that the name is a valid parameter
                msg = "'{}' is not a valid parameter for libsvm".format(k)
                raise mdp.NodeException(msg)
                
            if hasattr(self.parameter, k):
                setattr(self.parameter, k, v)
            else:
                msg = "'svm_parameter' has no attribute {}".format(k)
                raise AttributeError(msg)
            

    def _get_supported_dtypes(self):
        """Return the list of dtypes selfupported by this node."""
        # Support only float64 because of external library
        return ('float64',)

    def set_classifier(self, classifier):
        """
        Sets the classifier.

        classifier -- A string with the name of the classifier which
                      should be used. Possible values are in
                      self.classifiers
        """
        if classifier.upper() in self.classifiers:
            self.parameter.svm_type = getattr(libsvmutil, classifier.upper())
        else:
            msg = "Classifier Type %s is unknown or not supported." % classifier
            raise TypeError(msg)

    def set_kernel(self, kernel):
        """
        Sets the kernel.

        kernel     -- A string with the name of the classifier which
                      should be used. Possible values are in
                      self.kernels
        """
        if kernel.upper() in self.kernels:
            self.parameter.kernel_type = getattr(libsvmutil, kernel.upper())
        else:
            msg = "Kernel Type %s is unknown or not supported." % kernel
            raise TypeError(msg)

    def _stop_training(self):
        super(LibSVMClassifier, self)._stop_training()
        self.normalizer = _LabelNormalizer(self.labels)
                
        labels = self.normalizer.normalize(self.labels.tolist())
        features = self.data

        # Call svm training method.
        prob = libsvmutil.svm_problem(labels, features.tolist())
        # Train
        self.model = libsvmutil.svm_train(prob, self.parameter)

    def _label(self, x):
        if isinstance(x, (list, tuple, numx.ndarray)):
            y = [0] * len(x)
            p_labs, p_acc, p_vals = libsvmutil.svm_predict(y, x.tolist(), self.model)
            
            return numx.array(p_labs)
        else:
            msg = "Data must be a sequence of vectors"
            raise mdp.NodeException(msg)

    def predict_probability(self, x):
        self._pre_execution_checks(x)
        if isinstance(x, (list, tuple, numx.ndarray)):
            return self._prob(x)
        else:
            return self._prob([x])

    def _prob(self, x):
        y = [0] * len(x)
        p_labs, p_acc, p_vals = libsvmutil.svm_predict(y, x.tolist(), self.model, "-b 1")
        labels = self.model.get_labels()
        return [dict(list(zip(labels, ps))) for ps in p_vals]

    def _train(self, x, labels):
        super(LibSVMClassifier, self)._train(x, labels)



"""
Thread based scheduler for distribution across multiple CPU cores.
"""
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import threading
import time
import pickle as pickle

from .scheduling import Scheduler, cpu_count

SLEEP_TIME = 0.1  # time spend sleeping when waiting for a thread to finish


class ThreadScheduler(Scheduler):
    """Thread based scheduler.

    Because of the GIL this only makes sense if most of the time is spend in
    numpy calculations (or some other external non-blocking C code) or for IO,
    but can be more efficient than ProcessScheduler because of the
    shared memory.
    """

    def __init__(self, result_container=None, verbose=False, n_threads=1,
                 copy_callable=True):
        """Initialize the scheduler.

        result_container -- ResultContainer used to store the results.
        verbose -- Set to True to get progress reports from the scheduler
            (default value is False).
        n_threads -- Number of threads used in parallel. If None (default)
            then the number of detected CPU cores is used.
        copy_callable -- Use deep copies of the task callable in the threads.
            This is for example required if some nodes are stateful during
            execution (e.g., a BiNode using the coroutine decorator).
        """
        super(ThreadScheduler, self).__init__(
                                            result_container=result_container,
                                            verbose=verbose)
        if n_threads:
            self._n_threads = n_threads
        else:
            self._n_threads = cpu_count()
        self._n_active_threads = 0
        self.copy_callable = copy_callable

    def _process_task(self, data, task_callable, task_index):
        """Add a task, if possible without blocking.

        It blocks when the maximum number of threads is reached (given by
        n_threads) or when the system is not able to start a new thread.
        """
        task_started = False
        while not task_started:
            if self._n_active_threads >= self._n_threads:
                # release lock for other threads and wait
                self._lock.release()
                time.sleep(SLEEP_TIME)
                self._lock.acquire()
            else:
                self._lock.release()
                task_callable = task_callable.fork()
                if self.copy_callable:
                    # create a deep copy of the task_callable,
                    # since it might not be thread safe 
                    # (but the fork is still required)
                    as_str = pickle.dumps(task_callable, -1)
                    task_callable = pickle.loads(as_str)
                try:
                    thread = threading.Thread(target=self._task_thread,
                                              args=(data, task_callable,
                                                    task_index))
                    thread.start()
                    task_started = True
                except Exception:
                    if self.verbose:
                        print ("unable to create new thread,"
                               " waiting 2 seconds...")
                    time.sleep(2)

    def _task_thread(self, data, task_callable, task_index):
        """Thread function which processes a single task."""
        result = task_callable(data)
        self._store_result(result, task_index)
        self._n_active_threads -= 1



from builtins import range
from builtins import object
import mdp
n = mdp.numx

import py.test

from bimdp import BiNode, MSG_ID_SEP, BiFlow, BiClassifier, binode_coroutine
from bimdp.nodes import (
    IdentityBiNode, SFABiNode, FDABiNode, SignumBiClassifier
)
from ._tools import JumpBiNode


class TestBiNode(object):

    def test_msg_parsing1(self):
        """Test the message parsing and recombination."""
        class TestBiNode(BiNode):
            def _execute(self, x, a, b, d):
                self.a = a
                self.b = b
                self.d = d
                return x, {"g": 15, "z": 3}
            @staticmethod
            def is_trainable(): return False
        binode = TestBiNode(node_id="test")
        b_key = "test" + MSG_ID_SEP + "b"
        d_key = "test" + MSG_ID_SEP + "d"
        msg = {"c": 12, b_key: 42, "a": 13, d_key: "bla"}
        _, msg = binode.execute(None, msg)
        assert "a" in msg
        assert b_key not in msg
        assert d_key not in msg
        assert binode.a == 13
        assert binode.b == 42
        assert binode.d == "bla"
        # test the message combination
        assert msg["g"] == 15
        assert msg["z"] == 3

    def test_msg_parsing2(self):
        """Test that an adressed argument is not found."""
        class TestBiNode(BiNode):
            def _execute(self, x, a, b):
                self.a = a
                self.b = b
            @staticmethod
            def is_trainable(): return False
        binode = TestBiNode(node_id="test")
        b_key = "test" + MSG_ID_SEP + "b"
        # check that the 'd' key which is not an arg gets removed
        d_key = "test" + MSG_ID_SEP + "d"
        msg = {"c": 12, b_key: 42, "a": 13, d_key: "bla"}
        _, out_msg = binode.execute(None, msg)
        assert d_key not in out_msg

    def test_msg_magic(self):
        """Test that the magic msg argument works."""
        class TestBiNode(BiNode):
            def _execute(self, x, a, msg, b):
                self.a = a
                self.b = b
                del msg["c"]
                msg["f"] = 1
                return x, msg
            @staticmethod
            def is_trainable(): return False
        binode = TestBiNode(node_id="test")
        b_key = "test" + MSG_ID_SEP + "b"
        msg = {"c": 12, b_key: 42, "a": 13}
        _, msg = binode.execute(None, msg)
        assert "a" in msg
        assert "c" not in msg  # was deleted in _execute
        assert msg["f"] == 1
        assert b_key not in msg
        assert binode.a == 13
        assert binode.b == 42

    def test_method_magic(self):
        """Test the magic method message key."""
        class TestBiNode(BiNode):
            def _test(self, x, a, b):
                self.a = a
                self.b = b
            @staticmethod
            def is_trainable(): return False
        binode = TestBiNode(node_id="test")
        b_key = "test" + MSG_ID_SEP + "b"
        msg = {"c": 12, "a": 13, b_key: 42, "method": "test"}
        binode.execute(None, msg)
        assert "a" in msg
        assert b_key not in msg
        assert binode.b == 42

    def test_target_magic(self):
        """Test the magic target message key."""
        class TestBiNode(BiNode):
            def _execute(self, x, a, b):
                self.a = a
                self.b = b
            @staticmethod
            def is_trainable(): return False
        binode = TestBiNode(node_id="test")
        b_key = "test" + MSG_ID_SEP + "b"
        target_key = "test" + MSG_ID_SEP + "target"
        msg = {"c": 12, b_key: 42, "a": 13, target_key: "test2"}
        result = binode.execute(None, msg)
        assert len(result) == 3
        assert result[2] == "test2"

    def test_inverse_magic1(self):
        """Test the magic inverse method argument."""
        class TestBiNode(BiNode):
            def _inverse(self, x, a, b):
                self.a = a
                self.b = b
                y = n.zeros((len(x), self.input_dim))
                return y
            @staticmethod
            def is_trainable(): return False
        binode = TestBiNode(node_id="test", input_dim=20, output_dim=10)
        b_key = "test" + MSG_ID_SEP + "b"
        msg = {"c": 12, "a": 13, b_key: 42, "method": "inverse"}
        x = n.zeros((5, binode.output_dim))
        result = binode.execute(x, msg)
        assert len(result) == 3
        assert result[2] == -1
        assert result[0].shape == (5, 20)

    def test_inverse_magic2(self):
        """Test overriding the magic inverse target."""
        class TestBiNode(BiNode):
            def _inverse(self, x, a, b):
                self.a = a
                self.b = b
                y = n.zeros((len(x), self.input_dim))
                return y, None, "test2"
            @staticmethod
            def is_trainable(): return False
        binode = TestBiNode(node_id="test", input_dim=20, output_dim=10)
        b_key = "test" + MSG_ID_SEP + "b"
        msg = {"c": 12, "a": 13, b_key: 42, "method": "inverse"}
        x = n.zeros((5, binode.output_dim))
        result = binode.execute(x, msg)
        assert result[2] == "test2"

    def test_stoptrain_result1(self):
        """Test that stop_result is handled correctly."""
        stop_result = ({"test": 0}, 1)
        bi_sfa_node = SFABiNode(stop_result=stop_result,
                                node_id="testing binode")
        assert bi_sfa_node.is_trainable()
        x = n.random.random((100,10))
        train_result = bi_sfa_node.train(x)
        assert train_result == None
        assert bi_sfa_node.is_training()
        result = bi_sfa_node.stop_training()
        assert result == (None,) + stop_result
        assert bi_sfa_node.input_dim == 10
        assert bi_sfa_node.output_dim == 10
        assert bi_sfa_node.dtype == "float64"

    def test_stoptrain_result2(self):
        """Test that stop_result is handled correctly for multiple phases."""
        stop_result = [({"test": 0}, 1), ({"test2": 0}, 2)]
        binode = FDABiNode(stop_result=stop_result,
                           node_id="testing binode")
        x = n.random.random((100,10))
        msg = {"labels": n.zeros(len(x))}
        binode.train(x, msg)
        result = binode.stop_training()
        assert result == (None,) + stop_result[0]
        binode.train(x, msg)
        result = binode.stop_training()
        assert result == (None,) + stop_result[1]

    def test_stop_training_execute(self):
        """Test the magic execute method argument for stop_training."""
        class TestBiNode(BiNode):
            def _train(self, x): pass
            def _execute(self, x, a):
                self.a = a
                self.x = x
                y = n.zeros((len(x), self.output_dim))
                return y

        binode = TestBiNode(input_dim=20, output_dim=10)
        x = n.ones((5, binode.input_dim))
        binode.train(x)
        msg = {"x": x, "a": 13, "method": "execute"}
        result = binode.stop_training(msg)
        assert n.all(binode.x == x)
        assert binode.x.shape == (5, binode.input_dim)
        assert binode.a == 13
        assert len(result) == 2
        assert result[0].shape == (5, binode.output_dim)
        assert not n.any(result[0])

    def test_stop_training_inverse(self):
        """Test the magic inverse method argument for stop_training."""
        class TestBiNode(BiNode):
            def _train(self, x): pass
            def _inverse(self, x, a):
                self.a = a
                self.x = x
                y = n.zeros((len(x), self.input_dim))
                return y

        binode = TestBiNode(input_dim=20, output_dim=10)
        binode.train(n.ones((5, binode.input_dim)))
        x = n.ones((5, binode.output_dim))
        msg = {"x": x, "a": 13, "method": "inverse"}
        result = binode.stop_training(msg)
        assert n.all(binode.x == x)
        assert binode.x.shape == (5, binode.output_dim)
        assert binode.a == 13
        assert len(result) == 3
        assert result[2] == -1
        assert result[0].shape == (5, binode.input_dim)
        assert not n.any(result[0])

    def test_flow_from_sum(self):
        """Test the special addition method for BiNode."""
        node1 = IdentityBiNode()
        node2 = mdp.Node()
        flow = node1 + node2
        assert type(flow) is BiFlow
        node2 = IdentityBiNode()
        flow = node1 + node2
        assert type(flow) is BiFlow
        assert len(flow) == 2
        node3 = IdentityBiNode()
        flow = node1 + node2 + node3
        assert type(flow) is BiFlow
        assert len(flow) == 3
        node4 = IdentityBiNode()
        flow = node4 + flow
        assert type(flow) is BiFlow
        assert len(flow) == 4


class TestBiClassifierNode(object):

    def test_biclassifier(self):
        """Test the BiClassifier base class."""
        class TestBiClassifier(BiClassifier):
            def _label(self, x):
                return "LABELS"
            def _prob(self, x):
                return "PROPS"
            @staticmethod
            def is_trainable():
                return False
        node = TestBiClassifier()
        x = n.empty((5,2))
        msg = {"return_labels": "test->",
               "return_probs": True}
        result = node.execute(x, msg)
        assert result[0] is x
        assert "labels" not in result[1]
        assert result[1]["probs"] == "PROPS"
        assert result[1][msg["return_labels"] + "labels"] == "LABELS"
        assert "rank" not in result[1]
        msg = {"return_labels": None}
        result = node.execute(x,msg)
        assert result[0] is x
        assert "labels" not in result[1]
        assert "prop" not in result[1]
        assert "rank" not in result[1]

    def test_autogen_biclassifier(self):
        """Test that the autogenerated classifiers work."""
        node = SignumBiClassifier()
        msg = {"return_labels": True}
        # taken from the SignumClassifier unittest
        x = n.array([[1, 2, -3, -4], [1, 2, 3, 4]])
        result = node.execute(x, msg)
        assert result[0] is x
        assert result[1]["labels"].tolist() == [-1, 1]


class TestIdentityBiNode(object):

    def test_idnode(self):
        """Test the IdentityBiNode.

        Instantiation is tested and it should perform like an id node, but
        accept msg arguments.
        """
        binode = IdentityBiNode(node_id="testing binode")
        x = n.random.random((10,5))
        msg = {"some array": n.random.random((10,3))}
        # see if msg causes no problem
        y, msg = binode.execute(x, msg)
        assert n.all(x==y)
        # see if missing msg causes problem
        y = binode.execute(x)
        assert n.all(x==y)


class TestJumpBiNode(object):

    def test_node(self):
        """Test the JumpBiNode."""
        train_results = [[(0, "t1")], [None], [(3, "t3")]]
        stop_train_results = [None, (5, "st2"), (6, "st3")]
        execute_results = [(None, {}), None, (None, {}, "et4")]
        jumpnode = JumpBiNode(train_results=train_results,
                              stop_train_results=stop_train_results,
                              execute_results=execute_results)
        x = n.random.random((2,2))
        assert jumpnode.is_trainable()
        # training
        rec_train_results = []
        rec_stop_train_results = []
        for _ in range(len(train_results)):
            rec_train_results.append([jumpnode.train(x)])
            jumpnode.bi_reset()
            rec_stop_train_results.append(jumpnode.stop_training())
            jumpnode.bi_reset()
        assert not jumpnode.is_training()
        assert rec_train_results == train_results
        assert rec_stop_train_results == rec_stop_train_results
        # execution
        rec_execute_results = []
        for _ in range(4):  # note that this is more then the execute_targets
            rec_execute_results.append(jumpnode.execute(x))
        execute_results[1] = x
        execute_results.append(x)
        assert (rec_execute_results == execute_results)
        assert jumpnode.loop_counter == 4


class TestBiNodeCoroutine(object):
    """Test the coroutine decorator and the related BiNode functionality."""

    def test_codecorator(self):
        """Test basic codecorator functionality."""

        class CoroutineBiNode(BiNode):

            @staticmethod
            def is_trainable():
                return False

            @binode_coroutine(["alpha", "beta"])
            def _execute(self, x, alpha):
                """Blabla."""
                x, alpha, beta = yield (x, {"alpha": alpha, "beta": 2},
                                        self.node_id)
                x, alpha, beta = yield (x, {"alpha": alpha+1, "beta": beta+2},
                                        self.node_id)
                yield x, {"alpha": alpha, "beta": beta}

        node = CoroutineBiNode(node_id="conode")
        flow = BiFlow([node])
        x = n.random.random((3,2))
        y, msg = flow.execute(x, {"alpha": 3})
        assert msg["alpha"] == 4
        assert msg["beta"] == 4
        assert node.execute.__doc__ == """Blabla."""

    def test_codecorator2(self):
        """Test codecorator functionality with StopIteration."""

        class CoroutineBiNode(BiNode):

            @staticmethod
            def is_trainable():
                return False

            @binode_coroutine(["alpha", "beta"])
            def _execute(self, x, alpha):
                x, alpha, beta = yield (x, {"alpha": alpha, "beta": 2},
                                        self.node_id)
                x, alpha, beta = yield (x, {"alpha": alpha+1, "beta": beta+2},
                                        self.node_id)
                raise StopIteration(x, {"alpha": alpha, "beta": beta})

        node = CoroutineBiNode(node_id="conode")
        flow = BiFlow([node])
        x = n.random.random((3,2))
        y, msg = flow.execute(x, {"alpha": 3})
        assert msg["alpha"] == 4
        assert msg["beta"] == 4

    def test_codecorator_defaults(self):
        """Test codecorator argument default values."""

        class CoroutineBiNode(BiNode):

            @staticmethod
            def is_trainable():
                return False

            @binode_coroutine(["alpha", "beta"], defaults=(7,8))
            def _execute(self, x):
                x, alpha, beta = yield (x, None, self.node_id)
                raise StopIteration(x, {"alpha": alpha, "beta": beta})

        node = CoroutineBiNode(node_id="conode")
        flow = BiFlow([node])
        x = n.random.random((3,2))
        y, msg = flow.execute(x)
        assert msg["alpha"] == 7
        assert msg["beta"] == 8

    def test_codecorator_no_iteration(self):
        """Test codecorator corner case with no iterations."""

        class CoroutineBiNode(BiNode):

            @staticmethod
            def is_trainable():
                return False

            @binode_coroutine()
            def _execute(self, x):
                # at least one yield must be in a coroutine
                if False:
                    yield None
                raise StopIteration(None, {"a": 1}, self.node_id)

        node1 = CoroutineBiNode()
        x = n.random.random((3,2))
        result = node1.execute(x)
        assert result == (None, {"a": 1}, None)

    def test_codecorator_reset1(self):
        """Test that codecorator correctly resets after termination."""

        class CoroutineBiNode(BiNode):

            @staticmethod
            def is_trainable():
                return False

            @binode_coroutine()
            def _execute(self, x, a, msg=None):
                # note that the a argument is required, drop message
                for _ in range(2):
                    x = yield x
                raise StopIteration(x)

        node1 = CoroutineBiNode()
        x = n.random.random((3,2))
        # this inits the coroutine, a argument is needed
        node1.execute(x, {"a": 2})
        node1.execute(x)
        node1.execute(x)
        assert node1._coroutine_instances == {}
        # couroutine should be reset, a argument is needed again
        py.test.raises(TypeError, node1.execute, x)

    def test_codecorator_reset2(self):
        """Test that codecorator correctly resets without yields."""

        class CoroutineBiNode(BiNode):

            @staticmethod
            def is_trainable():
                return False

            @binode_coroutine()
            def _execute(self, x, a, msg=None):
                if False:
                    yield
                raise StopIteration(x)

        node1 = CoroutineBiNode()
        x = n.random.random((3,2))
        node1.execute(x, {"a": 2})
        assert node1._coroutine_instances == {}



from ._tools import *

TESTDECIMALS = {numx.dtype('d'): 12,
                numx.dtype('f'): 3,
                numx.dtype('D'): 12,
                numx.dtype('F'): 3,
                }

def test_eigenproblem(dtype, range, func):
    """Solve a standard eigenvalue problem."""
    dtype = numx.dtype(dtype)
    dim = 5
    if range:
        range = (2, dim -1)
    else:
        range = None
    a = utils.symrand(dim, dtype)+numx.diag([2.1]*dim).astype(dtype)
    w,z = func(a, range=range)
    # assertions
    assert_type_equal(z.dtype, dtype)
    w = w.astype(dtype)
    diag = numx.diagonal(utils.mult(utils.hermitian(z),
                                    utils.mult(a, z))).real
    assert_array_almost_equal(diag, w, TESTDECIMALS[dtype])

def test_geneigenproblem(dtype, range, func):
    """Solve a generalized eigenvalue problem."""
    dtype = numx.dtype(dtype)
    dim = 5
    if range:
        range = (2, dim -1)
    else:
        range = None
    a = utils.symrand(dim, dtype)
    b = utils.symrand(dim, dtype)+numx.diag([2.1]*dim).astype(dtype)
    w,z = func(a,b,range=range)
    # assertions
    assert z.dtype == dtype
    w = w.astype(dtype)
    diag1 = numx.diagonal(utils.mult(utils.hermitian(z),
                                     utils.mult(a, z))).real
    assert_array_almost_equal(diag1, w, TESTDECIMALS[dtype])
    diag2 = numx.diagonal(utils.mult(utils.hermitian(z),
                                     utils.mult(b, z))).real
    assert_array_almost_equal(diag2, numx.ones(diag2.shape[0]),
                              TESTDECIMALS[dtype])

test_geneigenproblem.funcs = [utils._symeig._symeig_fake]
if mdp.utils.symeig is utils._symeig.wrap_eigh:
    test_geneigenproblem.funcs.append(utils._symeig.wrap_eigh)

test_eigenproblem.funcs = test_geneigenproblem.funcs + [utils.nongeneral_svd]

def pytest_generate_tests(metafunc):
    for testtype in ('d', 'f'):
        for therange in (False, True):
            for func in metafunc.function.funcs:
                funcargs = dict(dtype=testtype,
                                range=therange,
                                func=func)
                theid = "%s, %s, %s" % (func.__name__, testtype, therange)
                metafunc.addcall(funcargs, id=theid)



from ._tools import *


def test_AdaptiveCutoffNode_smalldata():
    """Test AdaptiveCutoffNode thoroughly on a small data set."""
    # values from 0.1 to 0.6 and 0.2 to 0.7
    x1 = numx.array([[0.1, 0.3], [0.3, 0.5], [0.5, 0.7]])
    x2 = numx.array([[0.4, 0.6], [0.2, 0.4], [0.6, 0.2]])
    x = numx.concatenate([x1, x2])
    node = mdp.nodes.AdaptiveCutoffNode(lower_cutoff_fraction= 0.2, # clip first
                                    upper_cutoff_fraction=0.4) # last two
    node.train(x1)
    node.train(x2)
    node.stop_training()
    assert numx.all(x == node.data_hist)
    # test bound values
    assert numx.all(node.lower_bounds == numx.array([0.2, 0.3]))
    assert numx.all(node.upper_bounds == numx.array([0.4, 0.5]))
    # test execute
    x_test = (numx.array([[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]]))
    x_clip = node.execute(x_test)
    x_goal = numx.array([[0.2, 0.3], [0.3, 0.4], [0.4, 0.5]])
    assert (x_clip == x_goal).all()

def test_AdaptiveCutoffNode_randomdata():
    """Test AdaptiveCutoffNode on a large random data."""
    node = mdp.nodes.AdaptiveCutoffNode(lower_cutoff_fraction= 0.2,
                                    upper_cutoff_fraction=0.4,
                                    hist_fraction=0.5)
    x1 = numx_rand.random((1000, 3))
    x2 = numx_rand.random((500, 3))
    x = numx.concatenate([x1, x2])
    node.train(x1)
    node.train(x2)
    node.stop_training()
    node.execute(x)




from builtins import range
from builtins import object
import py.test

import mdp
from mdp import numx as np

from bimdp import (
    MessageResultContainer, BiFlow, BiFlowException, EXIT_TARGET, nodes,
)
from ._tools import TraceJumpBiNode, IdNode


class TestMessageResultContainer(object):
    """Test the behavior of the BetaResultContainer."""

    def test_mixed_dict(self):
        """Test msg being a dict containing an array."""
        rescont = MessageResultContainer()
        msg1 = {
            "f": 2,
            "a": np.zeros((10,3), 'int'),
            "b": "aaa",
            "c": 1,
        }
        msg2 = {
            "a": np.ones((15,3), 'int'),
            "b": "bbb",
            "c": 3,
            "d": 1,
        }
        rescont.add_message(msg1)
        rescont.add_message(msg2)
        combined_msg = rescont.get_message()
        a = np.zeros((25,3), 'int')
        a[10:] = 1
        reference_msg = {"a": a, "c": 4, "b": "aaabbb", "d": 1, "f": 2}
        assert np.all(reference_msg["a"] == reference_msg["a"])
        combined_msg.pop("a")
        reference_msg.pop("a")
        assert combined_msg == reference_msg

    def test_none_msg(self):
        """Test with one message being None."""
        rescont = MessageResultContainer()
        msgs = [None, {"a": 1}, None, {"a": 2, "b": 1}, None]
        for msg in msgs:
            rescont.add_message(msg)
        msg = rescont.get_message()
        assert msg == {"a": 3, "b": 1}

    def test_incompatible_arrays(self):
        """Test with incompatible arrays."""
        rescont = MessageResultContainer()
        msgs = [{"a":  np.zeros((10,3))}, {"a":  np.zeros((10,4))}]
        for msg in msgs:
            rescont.add_message(msg)
        py.test.raises(ValueError, rescont.get_message)


class TestBiFlow(object):

    def test_normal_flow(self):
        """Test a BiFlow with normal nodes."""
        flow = BiFlow([mdp.nodes.SFANode(output_dim=5),
                       mdp.nodes.PolynomialExpansionNode(degree=3),
                       mdp.nodes.SFANode(output_dim=20)])
        data_iterables = [[np.random.random((20,10)) for _ in range(6)],
                          None,
                          [np.random.random((20,10)) for _ in range(6)]]
        flow.train(data_iterables)
        x = np.random.random([100,10])
        flow.execute(x)

    def test_index_with_node_ids(self):
        """Test a BiFlow indexed by keys."""
        pca_node = nodes.PCABiNode(node_id="pca")
        biflow = BiFlow([pca_node])
        x = biflow["pca"]
        assert x is pca_node
        assert pca_node in biflow
        assert 'pca' in biflow

    def test_normal_multiphase(self):
        """Test training and execution with multiple training phases.

        The node with multiple training phases is a hinet.FlowNode.
        """
        sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
        sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
        flownode = mdp.hinet.FlowNode(mdp.Flow([sfa_node, sfa2_node]))
        flow = BiFlow([flownode,
                       mdp.nodes.PolynomialExpansionNode(degree=2),
                       mdp.nodes.SFANode(output_dim=5)])
        data_iterables = [[np.random.random((30,10)) for _ in range(6)],
                          None,
                          [np.random.random((30,10)) for _ in range(6)]]
        flow.train(data_iterables)
        x = np.random.random([100,10])
        flow.execute(x)

    def test_fda_binode(self):
        """Test using the FDABiNode in a BiFlow."""
        samples = mdp.numx_rand.random((100,10))
        labels = mdp.numx.arange(100)
        flow = BiFlow([mdp.nodes.PCANode(), nodes.FDABiNode()])
        flow.train([[samples],[samples]], [None,[{"labels": labels}]])

    def test_wrong_argument_handling(self):
        """Test correct error for additional arguments in Node instance."""
        samples = mdp.numx_rand.random((100,10))
        labels = mdp.numx.arange(100)
        # labels argument of FDANode is not supported in biflow
        flow = BiFlow([mdp.nodes.PCANode(), mdp.nodes.FDANode()])
        # the iterables are passed as if this were a normal Flow
        py.test.raises(BiFlowException,
                       flow.train, [[samples], [samples, labels]])
        # messing up the data iterables further doesn't matter, this is
        # actually interpreted as three data chunks for the FDANode training,
        # since argument iterables are not supported by BiFlow
        py.test.raises(BiFlowException,
                       flow.train, [[samples], [samples, labels, labels]])

    def test_training_targets(self):
        """Test targeting during training and stop_training."""
        tracelog = []
        verbose = False
        node1 = TraceJumpBiNode(
                    output_dim=1,
                    tracelog=tracelog,
                    node_id="node_1",
                    train_results=[[None]],
                    stop_train_results=[None],
                    execute_results=[None, (None, {"b": 2}, "node_3"),
                                     (None, {"b": 2}, EXIT_TARGET),],
                    verbose=verbose)
        node2 = TraceJumpBiNode(
                    output_dim=1,
                    tracelog=tracelog,
                    node_id="node_2",
                    train_results=[[None]],
                    stop_train_results=[(None, {"b": 2}, "node_1")],
                    execute_results=[None, (None, None, "node_1"),
                                     (None, None, "node_1")],
                    verbose=verbose)
        node3 = TraceJumpBiNode(
                    output_dim=1,
                    tracelog=tracelog,
                    node_id="node_3",
                    train_results=[[(None, {"a": 1}, "node_2"), None]],
                    stop_train_results=[(None, {"a": 1}, "node_2")],
                    execute_results=[(None, {"b": 2}, EXIT_TARGET)],
                    verbose=verbose)
        biflow = BiFlow([node1, node2, node3])
        data_iterables = [[np.random.random((1,1)) for _ in range(2)],
                          [np.random.random((1,1)) for _ in range(2)],
                          [np.random.random((1,1)) for _ in range(2)]]
        biflow.train(data_iterables)
        # print ",\n".join(str(log) for log in tracelog)
        # tracelog reference
        reference = [
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_1', 'train'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_1', 'train'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_1', 'stop_training'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_1', 'execute'),
            ('node_2', 'train'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_1', 'execute'),
            ('node_2', 'train'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_2', 'stop_training'),
            ('node_1', 'execute'),
            ('node_2', 'execute'),
            ('node_3', 'execute'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_1', 'execute'),
            ('node_2', 'execute'),
            ('node_3', 'train'),
            ('node_2', 'execute'),
            ('node_1', 'execute'),
            ('node_3', 'train'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_1', 'execute'),
            ('node_2', 'execute'),
            ('node_3', 'train'),
            ('node_2', 'execute'),
            ('node_1', 'execute'),
            ('node_3', 'train'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_3', 'stop_training'),
            ('node_2', 'execute'),
            ('node_3', 'execute'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset')
        ]
        assert tracelog == reference

    def test_execute_jump(self):
        """Test jumping around during execution."""
        tracelog = []
        verbose = False
        node1 = TraceJumpBiNode(
                    tracelog=tracelog,
                    node_id="node_1",
                    execute_results=[(None, None, "node_3"),
                                     (None, None, "node_2")],
                    verbose=verbose)
        node2 = TraceJumpBiNode(
                    tracelog=tracelog,
                    node_id="node_2",
                    execute_results=[(None, None, "node_1")],
                    verbose=verbose)
        node3 = TraceJumpBiNode(
                    tracelog=tracelog,
                    node_id="node_3",
                    execute_results=[(None, None, "node_1")],
                    verbose=verbose)
        biflow = BiFlow([node1, node2, node3])
        biflow.execute(None, {"a": 1})
        # bimdp.show_execution(biflow, x=None, msg={"a": 1}, debug=True)
        # tracelog reference
        reference = [
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
            ('node_1', 'execute'),
            ('node_3', 'execute'),
            ('node_1', 'execute'),
            ('node_2', 'execute'),
            ('node_1', 'execute'),
            ('node_2', 'execute'),
            ('node_3', 'execute'),
            ('node_1', 'bi_reset'),
            ('node_2', 'bi_reset'),
            ('node_3', 'bi_reset'),
        ]
        assert tracelog == reference

    def test_msg_normal_node(self):
        """Test that the msg is passed over a normal node."""
        node = IdNode()
        biflow = BiFlow([node])
        msg = {"a": 1}
        result = biflow.execute(np.random.random((1,1)), msg)
        assert msg == result[1]

    def test_exit_target(self):
        """Test that the magic exit target works."""
        tracelog = []
        node1 = TraceJumpBiNode(
                    tracelog=tracelog,
                    execute_results=[(None, None, EXIT_TARGET)],
                    verbose=False)
        node2 = IdNode()
        biflow = BiFlow([node1, node2])
        biflow.execute(None, {"a": 1})
        # bimdp.show_execution(biflow, x=None, msg={"a": 1}, debug=True)
        reference = [
           (None, 'bi_reset'), (None, 'execute'), (None, 'bi_reset')
        ]
        assert tracelog == reference

    def test_append_node_copy(self):
        """Test that appending a node does not perform a deept copy."""
        node1 = nodes.IdentityBiNode()
        node2 = nodes.IdentityBiNode()
        flow = BiFlow([node1])
        flow += node2
        assert flow[0] is node1
        assert type(flow) is BiFlow





import mdp
from ._tools import *

def testEtaComputerNode():
    tlen = 1e5
    t = numx.linspace(0,2*numx.pi,tlen)
    inp = numx.array([numx.sin(t), numx.sin(5*t)]).T
    # create node to be tested
    ecnode = mdp.nodes.EtaComputerNode()
    ecnode.train(inp)
    #
    etas = ecnode.get_eta(t=tlen)
    # precision gets better with increasing tlen
    assert_array_almost_equal(etas, [1, 5], decimal=4)




"""MDP extension to cache the execution phase of nodes.

This extension is based on the **joblib** library by Gael Varoquaux,
available at http://packages.python.org/joblib/. At the moment, the
extension is based on joblib v. 0.4.6.
"""
from builtins import object
__docformat__ = "restructuredtext en"

import joblib

from ..utils import TemporaryDirectory
from ..extension import ExtensionNode, activate_extension, deactivate_extension
from ..signal_node import Node

# -- global attributes for this extension

_cachedir = None
# If a temporary directory is used, a reference to the
# TemporaryDirectory object is kept here. The directory will be
# deleted when this object is destroyed, so either when this module is
# destroyed, or when a new directory is set and it is temporary
# directory again.
_cacheobj = None
# instance of joblib cache object (set with set_cachedir)
_memory = None

# True is the cache is active for *all* classes
_cache_active_global = True
_cached_classes = []
_cached_instances = []
_cached_methods = {}

def set_cachedir(cachedir=None, verbose=0):
    """Set root directory for the joblib cache.

    :Parameters:
     cachedir
         the cache directory name; if ``None``, a temporary directory
         is created using `TemporaryDirectory`
     verbose
         an integer number, controls the verbosity of the cache
         (default is 0, i.e., not verbose)
    """

    global _cachedir
    global _cacheobj
    global _cached_methods
    global _memory

    if cachedir is None:
        _cacheobj = TemporaryDirectory(prefix='mdp-joblib-cache.')
        cachedir = _cacheobj.name

    # only reset if the directory changes
    if cachedir != _cachedir:
        _cachedir = cachedir
        _memory = joblib.Memory(cachedir, verbose=verbose)
        # reset cached methods
        _cached_methods.clear()

# initialize cache with temporary directory
#set_cachedir()

class CacheExecuteExtensionNode(ExtensionNode, Node):
    """MDP extension for caching execution results.

    The return value of the 'execute' methods are cached if:
    1) the extension is activated in global mode
    2) the Node subclass is registered to be cached
    or
    3) the instance is registered to be cached

    *Warning: this extension might break the algorithms if nodes rely
    on side effects.*

    See `activate_caching`, `deactivate_caching`, and the `cache` context
    manager to learn about how to activate the caching mechanism and its
    options.
    """

    extension_name = 'cache_execute'

    def is_cached(self):
        """Return True if the node is cached."""
        global _cache_active_global
        global _cached_classes
        global _cached_instances
        return (_cache_active_global
                or self.__class__ in _cached_classes
                or self in _cached_instances)

    def set_instance_cache(self, active=True):
        """Add or remove this instance from caching.

        The global caching and class caching options still have priority over
        the instance caching option.
        """
        # add to global dictionary
        global _cached_instances
        if active:
            _cached_instances.append(self)
        else:
            if self in _cached_instances:
                _cached_instances.remove(self)

    def execute(self, x, *args, **kwargs):
        global _cached_methods

        # cache is not active for globally, for this class or instance:
        # call original execute method
        if not self.is_cached():
            return self._non_extension_execute(x, *args, **kwargs)

        if self not in _cached_methods:
            global _memory
            _cached_methods[self] = _memory.cache(
                self._non_extension_execute.__func__)
            # execute pre-execution checks once so that all automatic
            # settings of things like dtype and input_dim are done, and
            # caching begins from first execution, not the second
            self._pre_execution_checks(x)

        return _cached_methods[self](self, x, *args, **kwargs)


# ------- helper functions and context manager

# TODO: check that classes and instances are Nodes

def activate_caching(cachedir=None,
                     cache_classes=None, cache_instances=None,
                     verbose=0):
    """Activate caching extension.

    By default, the cache is activated globally (i.e., for all instances
    of Node). If cache_classes or cache instances are specified, the cache
    is activated only for those classes and instances.

    :Parameters:
     cachedir
      The root of the joblib cache, or a temporary directory if None
     cache_classes
      A list of Node subclasses for which caching is activated.
      Default value: None
     cache_classes
      A list of Node instances for which caching is activated.
      Default value: None
    """
    global _cache_active_global
    global _cached_classes
    global _cached_instances

    set_cachedir(cachedir=cachedir, verbose=verbose)
    _cache_active_global = (cache_classes is None and cache_instances is None)

    # active cache for specific classes and instances
    if cache_classes is not None:
        _cached_classes = list(cache_classes)
    if cache_instances is not None:
        _cached_instances = list(cache_instances)

    activate_extension('cache_execute')

def deactivate_caching(cachedir=None):
    """De-activate caching extension."""
    deactivate_extension('cache_execute')

    # reset global variables
    global _cache_active_global
    global _cached_classes
    global _cached_instances
    global _cached_methods
    _cache_active_global = True
    _cached_classes = []
    _cached_instances = []
    _cached_methods = {}

class cache(object):
    """Context manager for the 'cache_execute' extension.

    This allows using the caching extension using a 'with'
    statement, as in:

    >>> with mdp.caching.cache(CACHEDIR):                    # doctest: +SKIP
    ...     # 'node' is executed caching the results in CACHEDIR
    ...     node.execute(x)

    If the argument to the context manager is not specified, caching is
    done in a temporary directory.
    """

    def __init__(self, cachedir=None, cache_classes=None, cache_instances=None,
                 verbose=0):
        """Activate caching extension.

        By default, the cache is activated globally (i.e., for all instances
        of Node). If cache_classes or cache instances are specified, the cache
        is activated only for those classes and instances.

        :Parameters:
         cachedir
          The root of the joblib cache, or a temporary directory if None
         cache_classes
          A list of Node subclasses for which caching is activated.
          Default value: None
         cache_classes
          A list of Node instances for which caching is activated.
          Default value: None
        """
        self.cachedir = cachedir
        self.cache_classes = cache_classes
        self.cache_instances = cache_instances
        self.verbose = verbose

    def __enter__(self):
        activate_caching(self.cachedir, self.cache_classes,
                         self.cache_instances, self.verbose)

    def __exit__(self, type, value, traceback):
        deactivate_caching()



import mdp

def pca(x, **kwargs):
    """Filters multidimensional input data through its principal components.

    Observations of the same variable are stored on rows, different variables
    are stored on columns.

    This is a shortcut function for the corresponding node `nodes.PCANode`. If any
    keyword arguments are specified, they are passed to its constructor.

    This is equivalent to ``mdp.nodes.PCANode(**kwargs)(x)``
    """
    return mdp.nodes.PCANode(**kwargs)(x)

def fastica(x, **kwargs):
    """Perform Independent Component Analysis on input data using the FastICA
    algorithm by Aapo Hyvarinen.

    Observations of the same variable are stored on rows, different variables
    are stored on columns.

    This is a shortcut function for the corresponding node `nodes.FastICANode`.
    If any keyword arguments are specified, they are passed to its constructor.

    This is equivalent to ``mdp.nodes.FastICANode(**kwargs)(x)``
    """
    return mdp.nodes.FastICANode(**kwargs)(x)



## import mdp.parallel as parallel
## from _tools import *

## from test_pp_local import requires_parallel_python

## remote_slaves = [("localhost", 2)]

## @requires_parallel_python
## def test_simple():
##     scheduler = parallel.pp_support.NetworkPPScheduler(
##         remote_slaves=remote_slaves,
##         timeout=60,
##         verbose=False)
##     # process jobs
##     for i in range(30):
##         scheduler.add_task(i, parallel.SqrTestCallable())
##     results = scheduler.get_results()
##     scheduler.shutdown()
##     # check result
##     results.sort()
##     results = numx.array(results)
##     assert numx.all(results[:6] == numx.array([0,1,4,9,16,25]))



from __future__ import division
from builtins import range
from builtins import object
from past.utils import old_div
import mdp.parallel as parallel
from ._tools import *

def test_PCANode():
    """Test Parallel PCANode"""
    precision = 6
    x = numx_rand.random([100,10])
    x_test = numx_rand.random([20,10])
    # set different variances (avoid numerical errors)
    x *= numx.arange(1,11)
    x_test *= numx.arange(1,11)
    pca_node = mdp.nodes.PCANode()
    parallel_pca_node = parallel.ParallelPCANode()
    chunksize = 25
    chunks = [x[i*chunksize : (i+1)*chunksize]
                for i in range(len(x)//chunksize)]
    for chunk in chunks:
        pca_node.train(chunk)
        forked_node = parallel_pca_node.fork()
        forked_node.train(chunk)
        parallel_pca_node.join(forked_node)
    assert_array_almost_equal(pca_node._cov_mtx._cov_mtx,
                              parallel_pca_node._cov_mtx._cov_mtx,
                              precision)
    pca_node.stop_training()
    y1 = pca_node.execute(x_test)
    parallel_pca_node.stop_training()
    y2 = parallel_pca_node.execute(x_test)
    assert_array_almost_equal(abs(y1), abs(y2), precision)

def test_SFANode():
    """Test Parallel SFANode"""
    precision = 6
    x = numx_rand.random([100,10])
    x_test = numx_rand.random([20,10])
    # set different variances (avoid numerical errors)
    x *= numx.arange(1,11)
    x_test *= numx.arange(1,11)
    sfa_node = mdp.nodes.SFANode()
    parallel_sfa_node = parallel.ParallelSFANode()
    chunksize = 25
    chunks = [x[i*chunksize : (i+1)*chunksize]
                for i in range(len(x)//chunksize)]
    for chunk in chunks:
        sfa_node.train(chunk)
        forked_node = parallel_sfa_node.fork()
        forked_node.train(chunk)
        parallel_sfa_node.join(forked_node)
    assert_array_almost_equal(sfa_node._cov_mtx._cov_mtx,
                              parallel_sfa_node._cov_mtx._cov_mtx,
                              precision)
    sfa_node.stop_training()
    y1 = sfa_node.execute(x_test)
    parallel_sfa_node.stop_training()
    y2 = parallel_sfa_node.execute(x_test)
    assert_array_almost_equal(abs(y1), abs(y2), precision)

def test_FDANode():
    """Test Parallel FDANode."""
    # this test code is an adaption of the FDANode test
    precision = 4
    mean1 = [0., 2.]
    mean2 = [0., -2.]
    std_ = numx.array([1., 0.2])
    npoints = 50000
    rot = 45
    # input data: two distinct gaussians rotated by 45 deg
    def distr(size):
        return numx_rand.normal(0, 1., size=(size)) * std_
    x1 = distr((npoints,2)) + mean1
    utils.rotate(x1, rot, units='degrees')
    x2 = distr((npoints,2)) + mean2
    utils.rotate(x2, rot, units='degrees')
    # labels
    cl1 = numx.ones((x1.shape[0],), dtype='d')
    cl2 = 2.*numx.ones((x2.shape[0],), dtype='d')
    flow = parallel.ParallelFlow([parallel.ParallelFDANode()])
    flow.train([[(x1, cl1), (x2, cl2)]], scheduler=parallel.Scheduler())
    fda_node = flow[0]
    assert fda_node.tlens[1] == npoints
    assert fda_node.tlens[2] == npoints
    m1 = numx.array([mean1])
    m2 = numx.array([mean2])
    utils.rotate(m1, rot, units='degrees')
    utils.rotate(m2, rot, units='degrees')
    assert_array_almost_equal(fda_node.means[1], m1, 2)
    assert_array_almost_equal(fda_node.means[2], m2, 2)
    y = flow.execute([x1, x2], scheduler=parallel.Scheduler())
    assert_array_almost_equal(numx.mean(y, axis=0), [0., 0.], precision)
    assert_array_almost_equal(numx.std(y, axis=0), [1., 1.], precision)
    assert_almost_equal(utils.mult(y[:,0], y[:,1].T), 0., precision)
    v1 = old_div(fda_node.v[:,0],fda_node.v[0,0])
    assert_array_almost_equal(v1, [1., -1.], 2)
    v1 = old_div(fda_node.v[:,1],fda_node.v[0,1])
    assert_array_almost_equal(v1, [1., 1.], 2)

def test_ParallelHistogramNode_nofraction():
    """Test HistogramNode with fraction set to 1.0."""
    node = parallel.ParallelHistogramNode()
    x1 = numx.array([[0.1, 0.2], [0.3, 0.5]])
    x2 = numx.array([[0.3, 0.6], [0.2, 0.1]])
    x = numx.concatenate([x1, x2])
    chunks = [x1, x2]
    for chunk in chunks:
        forked_node = node.fork()
        forked_node.train(chunk)
        node.join(forked_node)
    assert numx.all(x == node.data_hist)
    node.stop_training()

def test_ParallelHistogramNode_fraction():
    """Test HistogramNode with fraction set to 0.5."""
    node = parallel.ParallelHistogramNode(hist_fraction=0.5)
    x1 = numx.random.random((1000, 3))
    x2 = numx.random.random((500, 3))
    chunks = [x1, x2]
    for chunk in chunks:
        forked_node = node.fork()
        forked_node.train(chunk)
        node.join(forked_node)
    assert len(node.data_hist) < 1000


class TestDerivedParallelMDPNodes(object):
    """Test derived nodes that use the parallel node classes."""

    def setup_method(self, method):
        if "parallel" in mdp.get_active_extensions():
            self.set_parallel = False
        else:
            mdp.activate_extension("parallel")
            self.set_parallel = True

    def teardown_method(self, method):
        if self.set_parallel:
            mdp.deactivate_extension("parallel")

    def test_WhiteningNode(self):
        """Test Parallel WhiteningNode"""
        x = numx_rand.random([100,10])
        x_test = numx_rand.random([20,10])
        # set different variances (avoid numerical errors)
        x *= numx.arange(1,11)
        x_test *= numx.arange(1,11)
        node = mdp.nodes.WhiteningNode()
        chunksize = 25
        chunks = [x[i*chunksize : (i+1)*chunksize]
                    for i in range(len(x)//chunksize)]
        for chunk in chunks:
            forked_node = node.fork()
            forked_node.train(chunk)
            node.join(forked_node)
        node.stop_training()
        node.execute(x_test)

    def test_SFA2Node(self):
        """Test Parallel SFA2Node"""
        x = numx_rand.random([100,10])
        x_test = numx_rand.random([20,10])
        # set different variances (avoid numerical errors)
        x *= numx.arange(1,11)
        x_test *= numx.arange(1,11)
        node = mdp.nodes.SFA2Node()
        chunksize = 25
        chunks = [x[i*chunksize : (i+1)*chunksize]
                    for i in range(len(x)//chunksize)]
        for chunk in chunks:
            forked_node = node.fork()
            forked_node.train(chunk)
            node.join(forked_node)
        node.stop_training()
        node.execute(x_test)




from .graph import ( Graph, GraphEdge, GraphException, GraphNode,
                    GraphTopologicalException, is_sequence,
                    recursive_map, recursive_reduce)

__all__ = ['Graph', 'GraphEdge', 'GraphException', 'GraphNode',
           'GraphTopologicalException', 'is_sequence',
           'recursive_map', 'recursive_reduce']

from mdp.utils import fixup_namespace
fixup_namespace(__name__, __all__,
                ('graph','fixup_namespace',))



import mdp
from ._tools import *

def testHitParadeNode():
    signal = uniform((300,3))
    gap = 5
    signal[10,0], signal[120,1], signal[230,2] = 4,3,2
    signal[11,0], signal[121,1], signal[231,2] = -4,-3,-2
    hit = mdp.nodes.HitParadeNode(1,gap,3)
    hit.train(signal[:100,:])
    hit.train(signal[100:200,:])
    hit.train(signal[200:300,:])
    maxima, max_ind = hit.get_maxima()
    minima, min_ind = hit.get_minima()
    assert_array_equal(maxima,numx.array([[4,3,2]]))
    assert_array_equal(max_ind,numx.array([[10,120,230]]))
    assert_array_equal(minima,numx.array([[-4,-3,-2]]))
    assert_array_equal(min_ind,numx.array([[11,121,231]]))
    # test integer type:
    signal = (uniform((300,3))*10).astype('i')
    gap = 5
    signal[10,0], signal[120,1], signal[230,2] = 40,30,20
    signal[11,0], signal[121,1], signal[231,2] = -40,-30,-20
    hit = mdp.nodes.HitParadeNode(1,gap,3)
    hit.train(signal[:100,:])
    hit.train(signal[100:200,:])
    hit.train(signal[200:300,:])
    maxima, max_ind = hit.get_maxima()
    minima, min_ind = hit.get_minima()
    assert_array_equal(maxima,numx.array([[40,30,20]]))
    assert_array_equal(max_ind,numx.array([[10,120,230]]))
    assert_array_equal(minima,numx.array([[-40,-30,-20]]))
    assert_array_equal(min_ind,numx.array([[11,121,231]]))

def testOneDimensionalHitParade():
    signal = (uniform(300)-0.5)*2
    gap = 5
    # put some maxima and minima
    signal[0] , signal[10] , signal[50] = 1.5, 1.4, 1.3
    signal[1] , signal[11] , signal[51] = -1.5, -1.4, -1.3
    # put two maxima and two minima within the gap
    signal[100], signal[103] = 2, 3
    signal[110], signal[113] = 3.1, 2
    signal[120], signal[123] = -2, -3.1
    signal[130], signal[133] = -3, -2
    hit = mdp.nodes._OneDimensionalHitParade(5,gap)
    hit.update((signal[:100],numx.arange(100)))
    hit.update((signal[100:200],numx.arange(100,200)))
    hit.update((signal[200:300],numx.arange(200,300)))
    maxima,ind_maxima = hit.get_maxima()
    minima,ind_minima = hit.get_minima()
    assert_array_equal(maxima,[3.1,3,1.5,1.4,1.3])
    assert_array_equal(ind_maxima,[110,103,0,10,50])
    assert_array_equal(minima,[-3.1,-3,-1.5,-1.4,-1.3])
    assert_array_equal(ind_minima,[123,130,1,11,51])



from __future__ import division
from builtins import str
from past.utils import old_div
import mdp
from mdp import numx, numx_linalg

class SymeigException(mdp.MDPException):
    pass

# the following functions and classes were part of the scipy_emulation.py file

_type_keys = ['f', 'd', 'F', 'D']
_type_conv = {('f','d'): 'd', ('f','F'): 'F', ('f','D'): 'D',
              ('d','F'): 'D', ('d','D'): 'D',
              ('F','d'): 'D', ('F','D'): 'D'}

def _greatest_common_dtype(alist):
    """
    Apply conversion rules to find the common conversion type
    dtype 'd' is default for 'i' or unknown types
    (known types: 'f','d','F','D').
    """
    dtype = 'f'
    for array in alist:
        if array is None:
            continue
        tc = array.dtype.char
        if tc not in _type_keys:
            tc = 'd'
        transition = (dtype, tc)
        if transition in _type_conv:
            dtype = _type_conv[transition]
    return dtype

def _assert_eigenvalues_real_and_positive(w, dtype):
    tol = numx.finfo(dtype.type).eps * 100
    if abs(w.imag).max() > tol:
        err = "Some eigenvalues have significant imaginary part: %s " % str(w)
        raise mdp.SymeigException(err)
    #if w.real.min() < 0:
    #    err = "Got negative eigenvalues: %s" % str(w)
    #    raise SymeigException(err)

def wrap_eigh(A, B = None, eigenvectors = True, turbo = "on", range = None,
              type = 1, overwrite = False):
    """Wrapper for scipy.linalg.eigh for scipy version > 0.7"""
    args = {}
    args['a'] = A
    args['b'] = B
    args['eigvals_only'] = not eigenvectors
    args['overwrite_a'] = overwrite
    args['overwrite_b'] = overwrite
    if turbo == "on":
        args['turbo'] = True
    else:
        args['turbo'] = False
    args['type'] = type
    if range is not None:
        n = A.shape[0]
        lo, hi = range
        if lo < 1:
            lo = 1
        if lo > n:
            lo = n
        if hi > n:
            hi = n
        if lo > hi:
            lo, hi = hi, lo
        # in scipy.linalg.eigh the range starts from 0
        lo -= 1
        hi -= 1
        range = (lo, hi)
    args['eigvals'] = range
    try:
        return numx_linalg.eigh(**args)
    except numx_linalg.LinAlgError as exception:
        raise SymeigException(str(exception))

def _symeig_fake(A, B = None, eigenvectors = True, turbo = "on", range = None,
                 type = 1, overwrite = False):
    """Solve standard and generalized eigenvalue problem for symmetric
(hermitian) definite positive matrices.
This function is a wrapper of LinearAlgebra.eigenvectors or
numarray.linear_algebra.eigenvectors with an interface compatible with symeig.

    Syntax:

      w,Z = symeig(A)
      w = symeig(A,eigenvectors=0)
      w,Z = symeig(A,range=(lo,hi))
      w,Z = symeig(A,B,range=(lo,hi))

    Inputs:

      A     -- An N x N matrix.
      B     -- An N x N matrix.
      eigenvectors -- if set return eigenvalues and eigenvectors, otherwise
                      only eigenvalues
      turbo -- not implemented
      range -- the tuple (lo,hi) represent the indexes of the smallest and
               largest (in ascending order) eigenvalues to be returned.
               1 <= lo < hi <= N
               if range = None, returns all eigenvalues and eigenvectors.
      type  -- not implemented, always solve A*x = (lambda)*B*x
      overwrite -- not implemented

    Outputs:

      w     -- (selected) eigenvalues in ascending order.
      Z     -- if range = None, Z contains the matrix of eigenvectors,
               normalized as follows:
                  Z^H * A * Z = lambda and Z^H * B * Z = I
               where ^H means conjugate transpose.
               if range, an N x M matrix containing the orthonormal
               eigenvectors of the matrix A corresponding to the selected
               eigenvalues, with the i-th column of Z holding the eigenvector
               associated with w[i]. The eigenvectors are normalized as above.
    """

    dtype = numx.dtype(_greatest_common_dtype([A, B]))
    try:
        if B is None:
            w, Z = numx_linalg.eigh(A)
        else:
            # make B the identity matrix
            wB, ZB = numx_linalg.eigh(B)
            _assert_eigenvalues_real_and_positive(wB, dtype)
            ZB = old_div(ZB.real, numx.sqrt(wB.real))
            # transform A in the new basis: A = ZB^T * A * ZB
            A = mdp.utils.mult(mdp.utils.mult(ZB.T, A), ZB)
            # diagonalize A
            w, ZA = numx_linalg.eigh(A)
            Z = mdp.utils.mult(ZB, ZA)
    except numx_linalg.LinAlgError as exception:
        raise SymeigException(str(exception))

    _assert_eigenvalues_real_and_positive(w, dtype)
    w = w.real
    Z = Z.real

    idx = w.argsort()
    w = w.take(idx)
    Z = Z.take(idx, axis=1)

    # sanitize range:
    n = A.shape[0]
    if range is not None:
        lo, hi = range
        if lo < 1:
            lo = 1
        if lo > n:
            lo = n
        if hi > n:
            hi = n
        if lo > hi:
            lo, hi = hi, lo

        Z = Z[:, lo-1:hi]
        w = w[lo-1:hi]

    # the final call to refcast is necessary because of a bug in the casting
    # behavior of Numeric and numarray: eigenvector does not wrap the LAPACK
    # single precision routines
    if eigenvectors:
        return mdp.utils.refcast(w, dtype), mdp.utils.refcast(Z, dtype)
    else:
        return mdp.utils.refcast(w, dtype)




from __future__ import print_function
from __future__ import division
from builtins import chr
from builtins import str
from builtins import range
from past.utils import old_div

from datetime import timedelta
import sys
import time

def get_termsize():
    """Return terminal size as a tuple (height, width)."""
    try:
        # this works on unix machines
        import struct, fcntl, termios
        height, width = struct.unpack("hhhh",
                                      fcntl.ioctl(0,termios.TIOCGWINSZ,
                                                  "\000"*8))[0:2]
        if not (height and width):
            height, width = 24, 79
    except ImportError:
        # for windows machins, use default values
        # Does anyone know how to get the console size under windows?
        # One approach is:
        # http://code.activestate.com/recipes/440694/
        height, width = 24, 79
    return height, width

def fmt_time(t, delimiters):
    """Return time formatted as a timedelta object."""
    meta_t = timedelta(seconds=round(t))
    return ''.join([delimiters[0], str(meta_t), delimiters[1]])

def _progress(percent, last, style, layout):
    # percentage string
    percent_s = "%3d%%" % int(round(percent*100))
    if style == 'bar':
        # how many symbols for such percentage
        symbols = int(round(percent * layout['width']))
        # build percent done arrow
        done = ''.join([layout['char1']*(symbols), layout['char2']])
        # build remaining space
        todo = ''.join([layout['char3']*(layout['width']-symbols)])
        # build the progress bar
        box =  ''.join([layout['delimiters'][0],
                        done, todo,
                        layout['delimiters'][1]])
        if layout['position'] == 'left':
            # put percent left
            box = ''.join(['\r', layout['indent'], percent_s, box])
        elif layout['position'] == 'right':
            # put percent right
            box = ''.join(['\r', layout['indent'], box, percent_s])
        else:
            # put it in the center
            percent_s = percent_s.lstrip()
            percent_idx = (len(box) // 2) - len(percent_s) + 2
            box = ''.join(['\r', layout['indent'],
                           box[0:percent_idx],
                           percent_s,
                           box[percent_idx+len(percent_s):]])
    else:
        now = time.time()
        if percent == 0:
            # write the time box directly
            tbox = ''.join(['?', layout['separator'], '?'])
        else:
            # Elapsed
            elapsed = now - layout['t_start']
            # Estimated total time
            if layout['speed'] == 'mean':
                e_t_a = old_div(elapsed,percent) - elapsed
            else:
                # instantaneous speed
                progress = percent-_progress.last_percent
                e_t_a = (1 - percent)/progress*(now-_progress.last_time)
            # build the time box
            tbox = ''.join([fmt_time(elapsed, layout['delimiters']),
                            layout['separator'],
                            fmt_time(e_t_a, layout['delimiters'])])
        # compose progress info box
        if layout['position'] == 'left':
            box = ''.join(['\r',
                           layout['indent'],
                           percent_s,
                           ' ',
                           tbox])
        else:
            box = ''.join(['\r',
                           layout['indent'],
                           tbox,
                           ' ',
                           percent_s])

        _progress.last_percent = percent
        _progress.last_time = now

    # print it only if something changed from last time
    if box != last:
        sys.stdout.write(box)
        sys.stdout.flush()
    return box

def progressinfo(sequence, length = None, style = 'bar', custom = None):
    """A fully configurable text-mode progress info box tailored to the
       command-line die-hards.

       To get a progress info box for your loops use it like this:

          >>> for i in progressinfo(sequence):
          ...     do_something(i)

      You can also use it with generators, files or any other iterable object,
      but in this case you have to specify the total length of the sequence:

          >>> for line in progressinfo(open_file, nlines):
          ...     do_something(line)

      If the number of iterations is not known in advance, you may prefer
      to iterate on the items directly. This can be useful for example if
      you are downloading a big file in a subprocess and want to monitor
      the progress. If the file to be downloaded is TOTAL bytes large and
      you are downloading it on local:
          >>> def done():
          ...     yield os.path.getsize(localfile)
          >>> for bytes in progressinfo(done(), -TOTAL)
          ...     time.sleep(1)
          ...     if download_process_has_finished():
          ...         break


     Arguments:

     sequence    - if it is a Python container object (list,
                   dict, string, etc...) and it supports the
                   __len__ method call, the length argument can
                   be omitted. If it is an iterator (generators,
                   file objects, etc...) the length argument must
                   be specified.

     Keyword arguments:

     length     - length of the sequence. Automatically set
                  if `sequence' has the __len__ method. If length is
                  negative, iterate on items.

     style      - If style == 'bar', display a progress bar. The
                  default layout is:

                  [===========60%===>.........]

                  If style == 'timer', display a time elapsed / time
                  remaining info box. The default layout is:

                  23% [02:01:28] - [00:12:37]

                  where fields have the following meaning:

                  percent_done% [time_elapsed] - [time_remaining]

     custom     - a dictionary for customizing the layout.
                  Default layout for the 'bar' style:
                   custom = { 'indent': '',
                              'width' : terminal_width - 1,
                              'position' : 'middle',
                              'delimiters' : '[]',
                              'char1' : '=',
                              'char2' : '>',
                              'char3' : '.' }


                  Default layout for the 'timer' style:
                   custom = { 'speed': 'mean',
                              'indent': '',
                              'position' : 'left',
                              'delimiters' : '[]',
                              'separator' : ' - ' }

                  Description:
                    speed = completion time estimation method, must be one of
                            ['mean', 'last']. 'mean' uses average speed, 'last'
                            uses last step speed.
                    indent = string used for indenting the progress info box
                    position = position of the percent done string,
                               must be one out of ['left', 'middle', 'right']


     Note 1: by default sys.stdout is flushed each time a new box is drawn.
             If you need to rely on buffered stdout you'd better not use this
             (any?) progress info box.
     Note 2: progressinfo slows down your loops. Always profile your scripts
             and check that you are not wasting 99% of the time in drawing
             the progress info box.
    """
    iterate_on_items = False
    # try to get the length of the sequence
    try:
        length = len(sequence)
    # if the object is unsized
    except TypeError:
        if length is None:
            err_str = "Must specify 'length' if sequence is unsized."
            raise Exception(err_str)
        elif length < 0:
            iterate_on_items = True
            length = -length
    length = float(length)
    # set layout
    if style == 'bar':
        layout = { 'indent': '',
                   'width' : get_termsize()[1],
                   'position' : 'middle',
                   'delimiters' : '[]',
                   'char1' : '=',
                   'char2' : '>',
                   'char3' : '.' }
        if custom is not None:
            layout.update(custom)
        fixed_lengths = len(layout['indent']) + 4
        if layout['position'] in ['left', 'right']:
            fixed_lengths += 4
        layout['width'] = layout['width'] - fixed_lengths
    elif style == 'timer':
        layout = { 'speed': 'mean',
                   'indent': '',
                   'position' : 'left',
                   'delimiters' : '[]',
                   'separator': ' - ',
                   't_start' : time.time()
                   }
        if custom is not None:
            layout.update(custom)
    else:
        err_str = "Style `%s' not known." % style
        raise ValueError(err_str)

    # start main loop
    last = None
    for count, value in enumerate(sequence):
        # generate progress info
        if iterate_on_items:
            last = _progress(old_div(value,length), last, style, layout)
        else:
            last = _progress(old_div(count,length), last, style, layout)
        yield value
    else:
        # we need this for the 100% notice
        if iterate_on_items:
            last = _progress(1., last, style, layout)
        else:
            last = _progress(old_div((count+1),length), last, style, layout)
    # clean up terminal
    sys.stdout.write('\n\r')

# execute this file for a demo of the progressinfo style
if __name__ == '__main__':
    #import random
    import mdp
    import tempfile
    print('Testing progressinfo...')
    # test various customized layouts
    cust_list = [ {'position' : 'left',
                   'indent': 'Progress: ',
                   'delimimters': '()',
                   'char3': ' '},
                  {},
                  {'position': 'right',
                   'width': 50} ]
    for cust in cust_list:
        test = 0
        for i in progressinfo(list(range(100, 600)), style = 'bar', custom = cust):
            test += i
            time.sleep(0.001)
        if test != 174750:
            raise Exception('Something wrong with progressinfo...')
    # generate random character sequence
    inp_list = []
    for j in range(500):
    #    inp_list.append(chr(random.randrange(256)))
        inp_list.append(chr(mdp.numx_rand.randint(256)))
    string = ''.join(inp_list)
    # test various customized layouts
    cust_list = [ {'position': 'left',
                   'separator': ' | ',
                   'delimiters': '()'},
                  {'position':'right'}]
    for cust in cust_list:
        out_list = []
        for i in progressinfo(string, style = 'timer', custom = cust):
            time.sleep(0.02)
            out_list.append(i)
        if inp_list != out_list:
            raise Exception('Something wrong with progressinfo...' )

    # write random file
    with tempfile.TemporaryFile(mode='r+') as fl:
        for i in range(1000):
            fl.write(str(i)+'\n')
        fl.flush()
        # rewind
        fl.seek(0)
        lines = []
        for line in progressinfo(fl, 1000):
            lines.append(int(line))
            time.sleep(0.01)
        if lines != list(range(1000)):
            raise Exception('Something wrong with progressinfo...' )

    # test iterate on items
    with tempfile.TemporaryFile(mode='r+') as fl:
        for i in range(10):
            fl.write(str(i)+'\n')
        fl.flush()
        # rewind
        fl.seek(0)
        def gen():
            for line_ in fl:
                yield int(line_)
        for line in progressinfo(gen(), -10, style='timer',
                                 custom={'speed':'last'}):
            time.sleep(1)

    print('Done.')



from builtins import range
import mdp
from ._tools import *

def test_VariadicCumulator():
    # create random data
    ONELEN = 101
    NREP = 7
    x = [numx_rand.rand(ONELEN, 3) for _ in range(NREP)]
    y = [numx_rand.rand(ONELEN, 3) for _ in range(NREP)]
    
    ABCumulator = mdp.VariadicCumulator('a', 'b')
    
    class TestABCumulator(ABCumulator):
        def _stop_training(self, *args, **kwargs):
            super(TestABCumulator, self)._stop_training(*args, **kwargs)
            # verify that the attributes are there
            assert hasattr(self, 'a')
            assert hasattr(self, 'b')
            # test tlen
            tlen = ONELEN*NREP
            assert self.tlen == tlen
            assert self.a.shape == (tlen, 3)
            assert self.b.shape == (tlen, 3)
            # test content
            for i in range(NREP):
                assert numx.all(self.a[i*ONELEN:(i+1)*ONELEN,:] == x[i])
                assert numx.all(self.b[i*ONELEN:(i+1)*ONELEN,:] == y[i])

    ab = TestABCumulator()
    for i in range(NREP):
        ab.train(x[i], y[i])
    ab.stop_training()
